import math

import torch
from torch.distributions import constraints

from .survival_distribution import SurvivalDistribution


class Normal(torch.distributions.Normal, SurvivalDistribution):
    def sf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        z = (value - self.loc) / self.scale
        return 1.0 - torch.special.ndtr(z)

    def logsf(self, value):
        return torch.log(self.sf(value))

    def logcdf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        z = (value - self.loc) / self.scale
        return torch.log(torch.special.ndtr(z))

    def isf(self, u):
        if self._validate_args:
            assert constraints.unit_interval.check(u).all()
        return self.loc + self.scale * torch.special.ndtri(1.0 - u)

    def _new_tensor(self, shape):
        return self.loc.new(shape)



from typing import Optional

import torch
from torch.distributions import Categorical

from .survival_distribution import SurvivalDistribution


class MixtureSameFamily(torch.distributions.MixtureSameFamily, SurvivalDistribution):
    def __init__(
        self, mixture_distribution, component_distribution, validate_args=None
    ):
        super(MixtureSameFamily, self).__init__(
            mixture_distribution=mixture_distribution,
            component_distribution=component_distribution,
            validate_args=validate_args,
        )

    def logsf(self, value):
        value = self._pad(value)
        log_sf_x = self.component_distribution.logsf(value)
        mix_logits = self.mixture_distribution.logits
        return torch.logsumexp(log_sf_x + mix_logits, dim=-1)

    def sf(self, value):
        value = self._pad(value)
        sf_x = self.component_distribution.sf(value)
        mix_prob = self.mixture_distribution.probs
        return torch.sum(sf_x * mix_prob, dim=-1)

    def sample_cond(
        self,
        sample_shape: torch.Size = torch.Size(),
        lower_bound: Optional[float] = None,
        upper_bound: Optional[float] = None,
    ):
        with torch.no_grad():
            sample_len = len(sample_shape)
            batch_len = len(self.batch_shape)
            gather_dim = sample_len + batch_len
            es = self.event_shape

            # Since we know that the sample lb < x < ub, we have to adjust the mixing
            # probabilities using the Bayes formula
            # p(z = k | lb < x < ub) \propto p(z = k) * Pr(lb < x < ub | z = k)
            dtype = self.mixture_distribution.probs.dtype
            device = self.mixture_distribution.probs.device
            if lower_bound is not None:
                lb = torch.as_tensor(lower_bound, dtype=dtype, device=device)
                above_lb_prob = self.component_distribution.sf(lb)
            else:
                above_lb_prob = 1.0

            if upper_bound is not None:
                ub = torch.as_tensor(upper_bound, dtype=dtype, device=device)
                above_ub_prob = self.component_distribution.sf(ub)
            else:
                above_ub_prob = 0.0

            # correction = Pr(lb < x < ub | z = k), shape [*batch_shape, num_components]
            correction = above_lb_prob - above_ub_prob
            conditional_mix_logits = (
                self.mixture_distribution.probs * correction
            ).log()
            mix_sample = Categorical(logits=conditional_mix_logits).sample(sample_shape)
            mix_shape = mix_sample.shape

            # component samples [n, B, k, E]
            comp_samples = self.component_distribution.sample_cond(
                sample_shape=sample_shape,
                lower_bound=lower_bound,
                upper_bound=upper_bound,
            )

            # Gather along the k dimension
            mix_sample_r = mix_sample.reshape(
                mix_shape + torch.Size([1] * (len(es) + 1))
            )
            mix_sample_r = mix_sample_r.repeat(
                torch.Size([1] * len(mix_shape)) + torch.Size([1]) + es
            )

            samples = torch.gather(comp_samples, gather_dim, mix_sample_r)
            return samples.squeeze(gather_dim)



from typing import Optional

import torch


class SurvivalDistribution(torch.distributions.Distribution):
    """Base class for all survival distributions.

    By default, all distributions are expected to implement logsf, log_prob and isf.
    Other methods (logcdf, log_hazard, rsample, rsample_cond) are derived from them.
    """

    def logsf(self, value):
        """Logarithm of the survival function.

        This is equal to the negative cumulative (integrated) hazard function.
        """
        raise NotImplementedError

    def sf(self, value):
        """Survival function.

        Also known as complementary cumulative distribution function (complementary CDF).

        References:
            https://en.wikipedia.org/wiki/Survival_function
        """
        return self.logsf(value).exp()

    def logcdf(self, value):
        """Logarithm of the cumulative distribution function."""
        return torch.log1p(-self.sf(value))

    def cdf(self, value):
        """Cumulative distribution function evaluated at value."""
        return 1.0 - self.sf(value)

    def isf(self, u):
        """Inverse of the survival function."""
        raise NotImplementedError

    def log_hazard(self, value):
        """Logarithm of the hazard function evaluated at value.

        Also known as failure rate.

        Hazard function is defined as PDF divided by the survival function.

        References:
            https://en.wikipedia.org/wiki/Survival_analysis#Hazard_function_and_cumulative_hazard_function
        """
        return self.log_prob(value) - self.logsf(value)

    def _new_tensor(self, shape):
        """Create new tensor with same dtype and device as distribution parameters."""
        raise NotImplementedError

    def rsample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        u = self._new_tensor(shape).uniform_()
        return self.isf(u)

    def rsample_cond(
        self,
        sample_shape: torch.Size = torch.Size(),
        lower_bound: Optional[float] = None,
        upper_bound: Optional[float] = None,
    ):
        """Draw a sample that is guaranteed to be between lower_bound and upper_bound.

        We generate a sample on the interval [upper_bound, upper_bound] as
            a = sf(upper_bound)
            b = sf(lower_bound)
            u ~ Uniform([a, b])  # note the switched order of a, b!
            x = isf(u)

        If neither lower_bound or upper_bound are given, this is equivalent to the
        standard rsample
            u ~ Uniform([0, 1])
            x = isf(u)
        """
        shape = self._extended_shape(sample_shape)
        # Sample u_full from Uniform([0, 1])
        u_full = self._new_tensor(shape).uniform_()
        if lower_bound is not None:
            lb = torch.as_tensor(lower_bound, dtype=u_full.dtype, device=u_full.device)
            u_max = self.sf(lb)
        else:
            u_max = 1.0
        if upper_bound is not None:
            ub = torch.as_tensor(upper_bound, dtype=u_full.dtype, device=u_full.device)
            u_min = self.sf(ub)
        else:
            u_min = 0.0
        # Equivalent to sampling u ~ Uniform([u_min, u_max])
        u = (u_max - u_min) * u_full + u_min
        # TODO: Should we detach u here?
        return self.isf(u)

    def sample_cond(
        self, sample_shape=torch.Size(), lower_bound=None, upper_bound=None
    ):
        with torch.no_grad():
            return self.rsample_cond(
                sample_shape=sample_shape,
                lower_bound=lower_bound,
                upper_bound=upper_bound,
            )



from numbers import Number
from typing import Union

import torch
from torch.distributions import constraints
from torch.distributions.utils import broadcast_all

from .survival_distribution import SurvivalDistribution


class Weibull(SurvivalDistribution):
    arg_constraints = {
        "rate": constraints.positive,
        "concentration": constraints.positive,
    }
    support = constraints.positive
    has_rsample = True

    def __init__(self, rate, concentration, validate_args=None):
        self.rate, self.concentration = broadcast_all(rate, concentration)
        batch_shape = self.rate.shape
        super().__init__(batch_shape, validate_args=validate_args)

    def logsf(self, x):
        if self._validate_args:
            self._validate_sample(x)
        return self.rate.neg() * torch.pow(x, self.concentration)

    def log_prob(self, x):
        return self.log_hazard(x) + self.logsf(x)

    def isf(self, u):
        if self._validate_args:
            assert constraints.unit_interval.check(u).all()
        return (-u.log() * self.rate.reciprocal()).pow(self.concentration.reciprocal())

    def log_hazard(self, x):
        if self._validate_args:
            self._validate_sample(x)
        return (
            self.rate.log()
            + self.concentration.log()
            + (self.concentration - 1) * x.log()
        )

    @property
    def mean(self):
        log_lmbd = self.concentration.reciprocal().neg() * self.rate.log()
        return torch.exp(log_lmbd + torch.lgamma(1 + self.concentration.reciprocal()))

    def _new_tensor(self, shape):
        return self.rate.new(shape)
